@@ -25,6 +25,8 @@ import com.strobel.core.VerifyArgument;
 
 import java.util.List;
 
+import static com.strobel.core.CollectionUtilities.lastOrDefault;
+
 /**
  * @author Mike Strobel
  */
@@ -99,6 +101,7 @@ public abstract class MetadataReader {
                 final int exceptionTableLength = buffer.readUnsignedShort();
                 final ExceptionTableEntry[] exceptionTable = new ExceptionTableEntry[exceptionTableLength];
 
+
                 for (int k = 0; k < exceptionTableLength; k++) {
                     final int startOffset = buffer.readUnsignedShort();
                     final int endOffset = buffer.readUnsignedShort();
@@ -177,12 +180,23 @@ public abstract class MetadataReader {
                     final String variableName = scope.lookupConstant(nameToken);
                     final String descriptor = scope.lookupConstant(typeToken);
 
+                    TypeReference parsedType;
+
+                    try {
+                        parsedType = getParser().parseTypeSignature(descriptor);
+                    }
+                    catch (final java.lang.Error | Exception ignored) {
+                        parsedType = null;
+                    }
+
                     entries[i] = new LocalVariableTableEntry(
                         variableIndex,
                         variableName,
-                        getParser().parseTypeSignature(descriptor),
+                        parsedType != null ? parsedType : BuiltinTypes.Object,
+                        descriptor,
                         scopeOffset,
-                        scopeLength
+                        scopeLength,
+                        parsedType == null
                     );
                 }
 
@@ -248,11 +262,23 @@ public abstract class MetadataReader {
                     final Object[] arguments = new Object[buffer.readUnsignedShort()];
                     final List<ParameterDefinition> parameters = bootstrapMethod.getParameters();
 
-                    if (parameters.size() != arguments.length + 3) {
+                    final int methodParameters = parameters.size();
+                    
+                    if (methodParameters != arguments.length + 3) {
                         final MethodDefinition resolved = bootstrapMethod.resolve();
 
-                        if (resolved == null || !resolved.isVarArgs() || parameters.size() >= arguments.length + 3) {
-                            throw Error.invalidBootstrapMethodEntry(bootstrapMethod, parameters.size(), arguments.length);
+                        ParameterDefinition lastParameter;
+
+                        final boolean varArgs = resolved != null ? resolved.isVarArgs()
+                                                                 : ((lastParameter = lastOrDefault(parameters)) != null &&
+                                                                    lastParameter.getParameterType().isArray());
+
+                        final int varArgsAdjustment = varArgs ? 1 : 0;
+
+                        final int varArgsAdjustedMethodParameters = methodParameters - varArgsAdjustment;
+
+                        if (varArgsAdjustedMethodParameters > arguments.length + 3) {
+                       		throw Error.invalidBootstrapMethodEntry(bootstrapMethod, varArgsAdjustedMethodParameters, arguments.length);
                         }
                     }
 
@@ -261,7 +287,7 @@ public abstract class MetadataReader {
                         final int token = buffer.readUnsignedShort();
                         final int parameterIndex = j + 3;
 
-                        if (parameterIndex < parameters.size()) {
+                        if (parameterIndex < methodParameters) {
                             parameterType = parameters.get(parameterIndex).getParameterType();
                         }
                         else {
