@@ -22,8 +22,10 @@ import com.strobel.assembler.ir.MetadataReader;
 import com.strobel.assembler.ir.attributes.*;
 import com.strobel.assembler.metadata.annotations.CustomAnnotation;
 import com.strobel.core.ArrayUtilities;
+import com.strobel.core.CollectionUtilities;
 import com.strobel.core.Comparer;
 import com.strobel.core.ExceptionUtilities;
+import com.strobel.core.Predicate;
 import com.strobel.core.StringUtilities;
 import com.strobel.core.VerifyArgument;
 import com.strobel.util.EmptyArrayCache;
@@ -75,7 +77,6 @@ public final class ClassFileReader extends MetadataReader {
 
         super();
 
-        _typeDefinition = new TypeDefinition();
         _options = options;
         _resolver = resolver;
         _resolverFrame = new ResolverFrame();
@@ -87,6 +88,11 @@ public final class ClassFileReader extends MetadataReader {
         _fields = new ArrayList<>();
         _methods = new ArrayList<>();
 
+        _typeDefinition = new TypeDefinition();
+        _typeDefinition.setResolver(_resolver);
+        _typeDefinition.setFlags(accessFlags);
+        _typeDefinition.setCompilerVersion(majorVersion, minorVersion);
+
         final int delimiter = _internalName.lastIndexOf('/');
 
         if (delimiter < 0) {
@@ -110,9 +116,6 @@ public final class ClassFileReader extends MetadataReader {
             _typeDefinition.setName(_internalName.substring(delimiterIndex + 1));
         }
 
-        _typeDefinition.setResolver(_resolver);
-        _typeDefinition.setFlags(accessFlags);
-        _typeDefinition.setCompilerVersion(majorVersion, minorVersion);
         _resolverFrame.addType(_typeDefinition);
         _parser = new MetadataParser(_typeDefinition);
         _scope = new Scope(_parser, _typeDefinition, constantPool);
@@ -255,11 +258,232 @@ public final class ClassFileReader extends MetadataReader {
 
                 return new InnerClassesAttribute(length, ArrayUtilities.asUnmodifiableList(entries));
             }
+
+            case AttributeNames.Record: {
+                // Record_attribute {
+                //     u2 attribute_name_index;
+                //     u4 attribute_length;
+                //     u2 components_count;
+                //     record_component_info components[components_count];
+                // }
+
+                final int componentCount = buffer.readUnsignedShort();
+                final RecordComponentInfo[] components = new RecordComponentInfo[componentCount];
+
+                // record_component_info {
+                //     u2 name_index;
+                //     u2 descriptor_index;
+                //     u2 attributes_count;
+                //     attribute_info attributes[attributes_count];
+                // }
+
+                for (int i = 0; i < components.length; i++) {
+                    final String componentName = _scope.lookupConstant(buffer.readUnsignedShort());
+                    final String componentDescriptor = _scope.lookupConstant(buffer.readUnsignedShort());
+                    final int componentAttributeCount = buffer.readUnsignedShort();
+                    final List<SourceAttribute> componentAttributes;
+
+                    TypeReference componentType;
+
+                    try {
+                        componentType = getParser().parseTypeSignature(componentDescriptor);
+                    }
+                    catch (final java.lang.Error | Exception ignored) {
+                        componentType = BuiltinTypes.Object;
+                    }
+
+                    if (componentAttributeCount > 0) {
+                        final SourceAttribute[] cAttr = new SourceAttribute[componentAttributeCount];
+
+                        for (int j = 0; j < cAttr.length; j++) {
+                            cAttr[j] = readAttribute(buffer);
+                        }
+
+                        componentAttributes = ArrayUtilities.asUnmodifiableList(cAttr);
+                    }
+                    else {
+                        componentAttributes = Collections.emptyList();
+                    }
+
+                    components[i] = new RecordComponentInfo(componentName, componentDescriptor, componentType, componentAttributes);
+                }
+
+                _scope._typeDefinition.setFlags(_scope._typeDefinition.getFlags() | Flags.RECORD);
+
+                return new RecordAttribute(length, ArrayUtilities.asUnmodifiableList(components));
+            }
+
+            case AttributeNames.Module: {
+                // Module_attribute {
+                //     u2 attribute_name_index;
+                //     u4 attribute_length;
+                //
+                //     u2 module_name_index;
+                //     u2 module_flags;
+                //     u2 module_version_index;
+                //
+                //     u2 requires_count; {...} requires[requires_count];
+                //     u2 exports_count; {...} exports[exports_count];
+                //     u2 opens_count; {...} opens[opens_count];
+                //     u2 uses_count; {...} uses[uses_count];
+                //     u2 provides_count; {...} provides[provides_count];
+                // }
+
+                final String moduleName = _scope.lookupConstant(buffer.readUnsignedShort());
+                final int flags = buffer.readUnsignedShort();
+                final int versionIndex = buffer.readUnsignedShort();
+                final String moduleVersion = versionIndex > 0 ? _scope.<String>lookupConstant(versionIndex) : null;
+
+                final int requiresCount = buffer.readUnsignedShort();
+                final ModuleDependency[] requires = requiresCount > 0 ? new ModuleDependency[requiresCount] : ModuleDependency.EMPTY;
+
+                for (int i = 0; i < requiresCount; i++) {
+                    requires[i] = readModuleDependency(buffer);
+                }
+
+                final int exportsCount = buffer.readUnsignedShort();
+                final PackageInfo[] exports = exportsCount > 0 ? new PackageInfo[exportsCount] : PackageInfo.EMPTY;
+
+                for (int i = 0; i < exportsCount; i++) {
+                    exports[i] = readPackageInfo(buffer);
+                }
+
+                final int opensCount = buffer.readUnsignedShort();
+                final PackageInfo[] opens = opensCount > 0 ? new PackageInfo[opensCount] : PackageInfo.EMPTY;
+
+                for (int i = 0; i < opensCount; i++) {
+                    opens[i] = readPackageInfo(buffer);
+                }
+
+                final int usesCount = buffer.readUnsignedShort();
+                final TypeReference[] uses = usesCount > 0 ? new TypeReference[usesCount] : TypeReference.EMPTY_REFERENCES;
+
+                for (int i = 0; i < usesCount; i++) {
+                    uses[i] = _scope.lookupType(buffer.readUnsignedShort());
+                }
+
+                final int providesCount = buffer.readUnsignedShort();
+                final ServiceInfo[] provides = providesCount > 0 ? new ServiceInfo[providesCount] : ServiceInfo.EMPTY;
+
+                for (int i = 0; i < providesCount; i++) {
+                    provides[i] = readServiceInfo(buffer);
+                }
+
+                return new ModuleAttribute(length, moduleName, moduleVersion, flags, requires, exports, opens, uses, provides);
+            }
+
+            case AttributeNames.ModulePackages: {
+                // ModulePackages_attribute {
+                //     u2 attribute_name_index;
+                //     u4 attribute_length;
+                //     u2 package_count;
+                //     u2 package_index[package_count];
+                // }
+
+                final int packageCount = buffer.readUnsignedShort();
+                final String[] packages = packageCount > 0 ? new String[packageCount] : EmptyArrayCache.EMPTY_STRING_ARRAY;
+
+                for (int i = 0; i < packageCount; i++) {
+                    packages[i] = _scope.lookupConstant(buffer.readUnsignedShort());
+                }
+
+                return new ModulePackagesAttribute(packages);
+            }
+
+            case AttributeNames.ModuleMainClass: {
+                // ModuleMainClass_attribute {
+                //     u2 attribute_name_index;
+                //     u4 attribute_length;
+                //     u2 main_class_index;
+                // }
+
+                return new ModuleMainClassAttribute(_scope.lookupType(buffer.readUnsignedShort()));
+            }
+
+            case AttributeNames.ModuleTarget: {
+                // ModuleTarget_attribute {
+                //     u2 attribute_name_index;
+                //     u4 attribute_length;
+                //     u2 target_index;
+                // }
+
+                return new ModuleTargetAttribute(_scope.<String>lookupConstant(buffer.readUnsignedShort()));
+            }
+
+            case AttributeNames.PermittedSubclasses: {
+                // PermittedSubclasses_attribute {
+                //     u2 attribute_name_index;
+                //     u4 attribute_length;
+                //     u2 number_of_classes;
+                //     u2 classes[number_of_classes];
+                // }
+
+                final TypeReference[] permittedSubclasses = new TypeReference[buffer.readUnsignedShort()];
+
+                for (int i = 0; i < permittedSubclasses.length; i++) {
+                    permittedSubclasses[i] = _scope.lookupType(buffer.readUnsignedShort());
+                }
+
+                return new PermittedSubclassesAttribute(permittedSubclasses);
+            }
         }
 
         return super.readAttributeCore(name, buffer, originalOffset, length);
     }
 
+    protected final ModuleDependency readModuleDependency(final Buffer buffer) {
+        // u2 requires_count;
+        // {   u2 requires_index;
+        //     u2 requires_flags;
+        //     u2 requires_version_index;
+        // } requires[requires_count];
+
+        final String name = _scope.lookupConstant(buffer.readUnsignedShort());
+        final int flags = buffer.readUnsignedShort();
+        final int versionIndex = buffer.readUnsignedShort();
+        final String version = versionIndex > 0 ? _scope.<String>lookupConstant(versionIndex) : null;
+
+        return new ModuleDependency(name, version, flags);
+    }
+
+    protected final PackageInfo readPackageInfo(final Buffer buffer) {
+        // u2 opens_count;
+        // {   u2 opens_index;
+        //     u2 opens_flags;
+        //     u2 opens_to_count;
+        //     u2 opens_to_index[opens_to_count];
+        // } opens[opens_count];
+
+        final String name = _scope.lookupConstant(buffer.readUnsignedShort());
+        final int flags = buffer.readUnsignedShort();
+        final int moduleCount = buffer.readUnsignedShort();
+        final String[] modules = moduleCount > 0 ? new String[moduleCount] : EmptyArrayCache.EMPTY_STRING_ARRAY;
+
+        for (int i = 0; i < moduleCount; i++) {
+            modules[i] = _scope.lookupConstant(buffer.readUnsignedShort());
+        }
+
+        return new PackageInfo(name, flags, modules);
+    }
+
+    protected final ServiceInfo readServiceInfo(final Buffer buffer) {
+        // u2 provides_count;
+        // {   u2 provides_index;
+        //     u2 provides_with_count;
+        //     u2 provides_with_index[provides_with_count];
+        // } provides[provides_count];
+
+        final TypeReference serviceInterface = _scope.lookupType(buffer.readUnsignedShort());
+        final int implementationCount = buffer.readUnsignedShort();
+        final TypeReference[] implementations = implementationCount > 0 ? new TypeReference[implementationCount] : TypeReference.EMPTY_REFERENCES;
+
+        for (int i = 0; i < implementationCount; i++) {
+            implementations[i] = _scope.lookupType(buffer.readUnsignedShort());
+        }
+
+        return new ServiceInfo(serviceInterface, implementations);
+    }
+
     @SuppressWarnings("ConstantConditions")
     private void readAttributesPhaseOne(final Buffer buffer, final SourceAttribute[] attributes) {
         for (int i = 0; i < attributes.length; i++) {
@@ -315,6 +539,11 @@ public final class ClassFileReader extends MetadataReader {
                     continue;
                 }
 
+                case AttributeNames.PermittedSubclasses: {
+                    attributes[i] = readAttributeCore(name, buffer, buffer.position(), length);
+                    continue;
+                }
+
                 default: {
                     final int offset = buffer.position();
                     final byte[] blob = new byte[length];
@@ -422,8 +651,8 @@ public final class ClassFileReader extends MetadataReader {
                         declaringMethod = null;
                     }
                 }
-                catch (final Throwable t) {
-                    throw ExceptionUtilities.asRuntimeException(t);
+                catch (final Exception e) {
+                    throw ExceptionUtilities.asRuntimeException(e);
                 }
 
                 if (declaringMethod != null) {
@@ -441,6 +670,7 @@ public final class ClassFileReader extends MetadataReader {
                     populateNamedInnerTypes();
                     populateAnonymousInnerTypes();
                     checkEnclosingMethodAttributes();
+                    checkModuleAttribute();
                 }
                 finally {
                     if (declaringMethod != null) {
@@ -459,6 +689,18 @@ public final class ClassFileReader extends MetadataReader {
         }
     }
 
+    private void checkModuleAttribute() {
+        final ModuleAttribute moduleAttribute = SourceAttribute.find(AttributeNames.Module, _attributes);
+
+        if (moduleAttribute != null &&
+            _typeDefinition.getDeclaredMethods().isEmpty() &&
+            _typeDefinition.getDeclaredFields().isEmpty() &&
+            _typeDefinition.getDeclaredTypes().isEmpty()) {
+
+            _typeDefinition.setFlags(_typeDefinition.getFlags() | Flags.MODULE);
+        }
+    }
+
     private void checkEnclosingMethodAttributes() {
         final InnerClassesAttribute innerClasses = SourceAttribute.find(AttributeNames.InnerClasses, _attributes);
 
@@ -671,6 +913,34 @@ public final class ClassFileReader extends MetadataReader {
         _typeDefinition.setBaseType(baseType);
     }
 
+    private void populatePermittedSubclasses() {
+        final PermittedSubclassesAttribute permittedSubclassesAttribute = SourceAttribute.find(AttributeNames.PermittedSubclasses, _attributes);
+
+        if (permittedSubclassesAttribute != null) {
+            _typeDefinition.setFlags(_typeDefinition.getFlags() | Flags.SEALED);
+            _typeDefinition.getPermittedSubclassesInternal().addAll(permittedSubclassesAttribute.getPermittedSubclasses());
+        }
+        else if (!_typeDefinition.isFinal()) {
+            final boolean unsealed = checkSealed0(_typeDefinition.getBaseType()) ||
+                                     CollectionUtilities.any(_typeDefinition.getExplicitInterfaces(),
+                                                             new Predicate<TypeReference>() {
+                                                                 @Override
+                                                                 public boolean test(final TypeReference type) {
+                                                                     return checkSealed0(type);
+                                                                 }
+                                                             });
+            if (unsealed) {
+                _typeDefinition.setFlags(_typeDefinition.getFlags() | Flags.NON_SEALED);
+            }
+        }
+    }
+
+    private boolean checkSealed0(final TypeReference type) {
+        final TypeDefinition resolved = type != null ? type.resolve() : null;
+        final boolean isSealed = resolved != null && resolved.isSealed();
+        return isSealed;
+    }
+
     private void populateNamedInnerTypes() {
         final InnerClassesAttribute innerClasses = SourceAttribute.find(AttributeNames.InnerClasses, _attributes);
 
@@ -1071,8 +1341,8 @@ public final class ClassFileReader extends MetadataReader {
                 }
             }
         }
-        catch (final Throwable t) {
-            throw ExceptionUtilities.asRuntimeException(t);
+        catch (final Exception e) {
+            throw ExceptionUtilities.asRuntimeException(e);
         }
     }
 
@@ -1153,6 +1423,8 @@ public final class ClassFileReader extends MetadataReader {
                 Collections.addAll(annotations, invisibleAnnotations.getAnnotations());
             }
         }
+
+        populatePermittedSubclasses();
     }
 
     // </editor-fold>
