@@ -61,6 +61,8 @@ public class CoreMetadataFactory implements MetadataFactory {
     public GenericParameter makeTypeVariable(final String name, final FieldTypeSignature[] bounds) {
         final GenericParameter genericParameter = new GenericParameter(name);
 
+        genericParameter.setResolver(_resolver);
+
         if (ArrayUtilities.isNullOrEmpty(bounds)) {
             return genericParameter;
         }
@@ -141,7 +143,7 @@ public class CoreMetadataFactory implements MetadataFactory {
 
         interfaceTypes = ArrayUtilities.asUnmodifiableList(it);
 
-        return new CompoundTypeReference(baseType, interfaceTypes);
+        return new CompoundTypeReference(baseType, interfaceTypes, _resolver);
     }
 
     public TypeReference makeParameterizedType(
