@@ -16,10 +16,12 @@
 
 package com.strobel.decompiler.ast;
 
+import com.strobel.assembler.metadata.ParameterDefinition;
 import com.strobel.core.Comparer;
 import com.strobel.core.Predicate;
 import com.strobel.core.StrongBox;
 import com.strobel.core.VerifyArgument;
+import com.strobel.functions.Consumer;
 import com.strobel.util.ContractUtils;
 
 import java.util.ArrayList;
@@ -43,23 +45,29 @@ public final class PatternMatching {
                match(node, AstCode.EndFinally);
     }
 
-    public static <T> boolean matchGetOperand(final Node node, final AstCode code, final StrongBox<? super T> operand) {
+    public static <T> boolean matchGetOperand(final Node node, final AstCode code, final Consumer<? super T> operand) {
         if (node instanceof Expression) {
             final Expression expression = (Expression) node;
 
             if (expression.getCode() == code &&
                 expression.getArguments().isEmpty()) {
 
-                operand.set(expression.getOperand());
+                final @SuppressWarnings("unchecked") T typedOperand = (T) expression.getOperand();
+                operand.accept(typedOperand);
                 return true;
             }
         }
 
-        operand.set(null);
+        operand.accept(null);
         return false;
     }
 
-    public static <T> boolean matchGetOperand(final Node node, final AstCode code, final Class<T> operandType, final StrongBox<? super T> operand) {
+    public static <T> boolean matchGetOperand(
+        final Node node,
+        final AstCode code,
+        final Class<T> operandType,
+        final StrongBox<? super T> operand) {
+
         if (node instanceof Expression) {
             final Expression expression = (Expression) node;
 
@@ -92,7 +100,12 @@ public final class PatternMatching {
         return false;
     }
 
-    public static <T> boolean matchGetArguments(final Node node, final AstCode code, final StrongBox<? super T> operand, final List<Expression> arguments) {
+    public static <T> boolean matchGetArguments(
+        final Node node,
+        final AstCode code,
+        final StrongBox<? super T> operand,
+        final List<Expression> arguments) {
+
         if (node instanceof Expression) {
             final Expression expression = (Expression) node;
 
@@ -306,15 +319,10 @@ public final class PatternMatching {
     public static boolean matchSimpleBreak(final BasicBlock block, final Label label) {
         final List<Node> body = block.getBody();
 
-        if (body.size() == 2 &&
-            body.get(0) instanceof Label &&
-            match(body.get(1), AstCode.Goto) &&
-            ((Expression) body.get(1)).getOperand() == label) {
-
-            return true;
-        }
-
-        return false;
+        return body.size() == 2 &&
+               body.get(0) instanceof Label &&
+               match(body.get(1), AstCode.Goto) &&
+               ((Expression) body.get(1)).getOperand() == label;
     }
 
     public static boolean matchAssignmentAndConditionalBreak(
@@ -346,17 +354,13 @@ public final class PatternMatching {
                 return true;
             }
 
-            if (match(e, AstCode.StoreElement)) {
-                assignedValue.set(e.getArguments().get(2));
-                Expression arg0 = e.getArguments().get(0).clone();
-                Expression arg1 = e.getArguments().get(1).clone();
-                equivalentLoad.set(new Expression(AstCode.LoadElement, null, arg0.getOffset(), arg0, arg1));
+            if (matchElementAssignment(e, assignedValue, equivalentLoad)) {
                 return true;
             }
 
             if (match(e, AstCode.PutField)) {
+                final Expression arg0 = e.getArguments().get(0).clone();
                 assignedValue.set(e.getArguments().get(1));
-                Expression arg0 = e.getArguments().get(0).clone();
                 equivalentLoad.set(new Expression(AstCode.GetField, null, arg0.getOffset(), arg0));
                 return true;
             }
@@ -389,7 +393,11 @@ public final class PatternMatching {
         return false;
     }
 
-    public static boolean matchAssignment(final Node node, final StrongBox<Expression> assignedValue, final StrongBox<Expression> equivalentLoad) {
+    public static boolean matchAssignment(
+        final Node node,
+        final StrongBox<Expression> assignedValue,
+        final StrongBox<Expression> equivalentLoad) {
+
         if (node instanceof Expression) {
             final Expression e = (Expression) node;
 
@@ -405,17 +413,13 @@ public final class PatternMatching {
                 return true;
             }
 
-            if (match(e, AstCode.StoreElement)) {
-                assignedValue.set(e.getArguments().get(2));
-                Expression arg0 = e.getArguments().get(0).clone();
-                Expression arg1 = e.getArguments().get(1).clone();
-                equivalentLoad.set(new Expression(AstCode.LoadElement, null, arg0.getOffset(), arg0, arg1));
+            if (matchElementAssignment(e, assignedValue, equivalentLoad)) {
                 return true;
             }
 
             if (match(e, AstCode.PutField)) {
+                final Expression arg0 = e.getArguments().get(0).clone();
                 assignedValue.set(e.getArguments().get(1));
-                Expression arg0 = e.getArguments().get(0).clone();
                 equivalentLoad.set(new Expression(AstCode.GetField, e.getOperand(), arg0.getOffset(), arg0));
                 return true;
             }
@@ -425,6 +429,28 @@ public final class PatternMatching {
         return false;
     }
 
+    private static boolean matchElementAssignment(
+        final Node node,
+        final StrongBox<Expression> assignedValue,
+        final StrongBox<Expression> equivalentLoad) {
+
+        if (match(node, AstCode.StoreElement)) {
+            final Expression e = (Expression) node;
+            final Expression a0 = e.getArguments().get(0).clone();
+            final Expression a1 = e.getArguments().get(1).clone();
+
+            assignedValue.set(e.getArguments().get(2));
+            equivalentLoad.set(new Expression(AstCode.LoadElement, null, a0.getOffset(), a0, a1));
+
+            return true;
+        }
+
+        assignedValue.set(null);
+        equivalentLoad.set(null);
+
+        return false;
+    }
+
     public static boolean matchLast(final BasicBlock block, final AstCode code) {
         final List<Node> body = block.getBody();
 
@@ -479,17 +505,7 @@ public final class PatternMatching {
         final StrongBox<? super T> operand,
         final StrongBox<Expression> argument) {
 
-        final List<Node> body = block.getBody();
-
-        if (body.size() >= 1 &&
-            matchGetArgument(body.get(body.size() - 1), code, operand, argument)) {
-
-            return true;
-        }
-
-        operand.set(null);
-        argument.set(null);
-        return false;
+        return matchLast(block.getBody(), code, operand, argument);
     }
 
     public static <T> boolean matchLast(
@@ -498,7 +514,14 @@ public final class PatternMatching {
         final StrongBox<? super T> operand,
         final StrongBox<Expression> argument) {
 
-        final List<Node> body = block.getBody();
+        return matchLast(block.getBody(), code, operand, argument);
+    }
+
+    private static <T> boolean matchLast(
+        final List<Node> body,
+        final AstCode code,
+        final StrongBox<? super T> operand,
+        final StrongBox<Expression> argument) {
 
         if (body.size() >= 1 &&
             matchGetArgument(body.get(body.size() - 1), code, operand, argument)) {
@@ -535,10 +558,11 @@ public final class PatternMatching {
 
     public static boolean matchThis(final Node node) {
         final StrongBox<Variable> operand = new StrongBox<>();
+        final ParameterDefinition p;
 
         return matchGetOperand(node, AstCode.Load, operand) &&
-               operand.get().isParameter() &&
-               operand.get().getOriginalParameter().getPosition() == -1;
+               (p = operand.get().getOriginalParameter()) != null &&
+               p.getPosition() == -1;
     }
 
     public static boolean matchLoadAny(final Node node, final Iterable<Variable> expectedVariables) {
@@ -553,15 +577,101 @@ public final class PatternMatching {
         );
     }
 
-    public static boolean matchLoad(final Node node, final StrongBox<Variable> variable) {
+    public static boolean matchLoad(final Node node, final Consumer<? super Variable> variable) {
         return matchGetOperand(node, AstCode.Load, variable);
     }
 
-    public static boolean matchStore(final Node node, final StrongBox<Variable> variable, final StrongBox<Expression> argument) {
+    public static boolean matchNumericLdC(final Node node, final StrongBox<? super Number> value) {
+        if (matchGetOperand(node, AstCode.LdC, value) &&
+            value.get() instanceof Number) {
+
+            return true;
+        }
+
+        value.set(null);
+        return false;
+    }
+
+    public static boolean matchVariableIncDec(final Node node, final StrongBox<Variable> variable) {
+        if (node instanceof Expression) {
+            final Expression e = (Expression) node;
+            final AstCode code = e.getCode();
+
+            if (code.isIncDec()) {
+                final Variable v;
+
+                @SuppressWarnings({ "unchecked", "rawtypes" })
+                final StrongBox<Number> valueBox = (StrongBox)variable;
+                final StrongBox<Expression> argument = new StrongBox<>();
+
+                if (matchGetArgument(e, AstCode.Inc, variable, argument) &&
+                    (v = variable.get()) != null &&
+                    matchNumericLdC(argument.get(), valueBox)) {
+
+                    variable.set(v);
+                    return true;
+                }
+
+                //noinspection ConstantConditions
+                if (matchGetArgument(e, code, valueBox, argument) &&
+                    valueBox.get() instanceof Number &&
+                    matchLoad(argument.get(), variable)) {
+
+                    return true;
+                }
+            }
+        }
+
+        variable.set(null);
+        return false;
+    }
+
+    public static boolean matchVariableIncDec(
+        final Node node,
+        final StrongBox<Variable> variable,
+        final StrongBox<Number> amount) {
+
+        if (node instanceof Expression) {
+            final Expression e = (Expression) node;
+            final AstCode code = e.getCode();
+
+            if (code.isIncDec()) {
+                final StrongBox<Expression> argument = new StrongBox<>();
+
+                if (matchGetArgument(e, AstCode.Inc, variable, argument) &&
+                    matchNumericLdC(argument.get(), amount)) {
+
+                    return true;
+                }
+
+                //noinspection ConstantConditions
+                if (matchGetArgument(e, code, amount, argument) &&
+                    amount.get() instanceof Number &&
+                    matchLoad(argument.get(), variable)) {
+
+                    return true;
+                }
+            }
+        }
+
+        variable.set(null);
+        amount.set(null);
+        return false;
+    }
+
+    public static boolean matchStore(
+        final Node node,
+        final StrongBox<Variable> variable,
+        final StrongBox<Expression> argument) {
+
         return matchGetArgument(node, AstCode.Store, variable, argument);
     }
 
-    public static boolean matchStore(final Node node, final StrongBox<Variable> variable, final List<Expression> argument) {
+    public static boolean matchStore(
+        final Node node,
+        final StrongBox<Variable> variable,
+        final List<Expression> argument) {
+
         return matchGetArguments(node, AstCode.Store, variable, argument);
     }
 
@@ -582,7 +692,11 @@ public final class PatternMatching {
                Comparer.equals(((Expression) node).getOperand(), expectedVariable);
     }
 
-    public static boolean matchStore(final Node node, final Variable expectedVariable, final StrongBox<Expression> value) {
+    public static boolean matchStore(
+        final Node node,
+        final Variable expectedVariable,
+        final StrongBox<Expression> value) {
+
         final StrongBox<Variable> v = new StrongBox<>();
 
         if (matchGetArgument(node, AstCode.Store, v, value) &&
@@ -596,14 +710,22 @@ public final class PatternMatching {
         return false;
     }
 
-    public static boolean matchLoad(final Node node, final Variable expectedVariable, final StrongBox<Expression> argument) {
+    public static boolean matchLoad(
+        final Node node,
+        final Variable expectedVariable,
+        final StrongBox<Expression> argument) {
+
         final StrongBox<Variable> operand = new StrongBox<>();
 
         return matchGetArgument(node, AstCode.Load, operand, argument) &&
                Comparer.equals(operand.get(), expectedVariable);
     }
 
-    public static boolean matchLoadStore(final Node node, final Variable expectedVariable, final StrongBox<Variable> targetVariable) {
+    public static boolean matchLoadStore(
+        final Node node,
+        final Variable expectedVariable,
+        final StrongBox<Variable> targetVariable) {
+
         final StrongBox<Expression> temp = new StrongBox<>();
 
         if (matchGetArgument(node, AstCode.Store, targetVariable, temp) &&
@@ -616,7 +738,11 @@ public final class PatternMatching {
         return false;
     }
 
-    public static boolean matchLoadStoreAny(final Node node, final Iterable<Variable> expectedVariables, final StrongBox<Variable> targetVariable) {
+    public static boolean matchLoadStoreAny(
+        final Node node,
+        final Iterable<Variable> expectedVariables,
+        final StrongBox<Variable> targetVariable) {
+
         for (final Variable variable : VerifyArgument.notNull(expectedVariables, "expectedVariables")) {
             if (matchLoadStore(node, variable, targetVariable)) {
                 return true;
@@ -626,7 +752,11 @@ public final class PatternMatching {
         return false;
     }
 
-    public static boolean matchBooleanComparison(final Node node, final StrongBox<Expression> argument, final StrongBox<Boolean> comparand) {
+    public static boolean matchBooleanComparison(
+        final Node node,
+        final StrongBox<Expression> argument,
+        final StrongBox<Boolean> comparand) {
+
         final List<Expression> a = new ArrayList<>(2);
 
         if (matchGetArguments(node, AstCode.CmpEq, a) || matchGetArguments(node, AstCode.CmpNe, a)) {
@@ -652,7 +782,11 @@ public final class PatternMatching {
         return false;
     }
 
-    public static boolean matchComparison(final Node node, final StrongBox<Expression> left, final StrongBox<Expression> right) {
+    public static boolean matchComparison(
+        final Node node,
+        final StrongBox<Expression> left,
+        final StrongBox<Expression> right) {
+
         if (node instanceof Expression) {
             final Expression e = (Expression) node;
 
@@ -737,6 +871,41 @@ public final class PatternMatching {
         return Boolean.FALSE.equals(matchBooleanConstant(node));
     }
 
+    public static boolean matchIntegralConstant(final Node node, final Consumer<? super Long> constant) {
+        if (match(node, AstCode.LdC)) {
+            final Object operand = ((Expression) node).getOperand();
+
+            if (operand instanceof Number && !(operand instanceof Float || operand instanceof Double)) {
+                constant.accept(((Number) operand).longValue());
+                return true;
+            }
+        }
+
+        return false;
+    }
+
+    public static boolean matchNumericConstant(final Node node, final Consumer<? super Number> constant) {
+        if (match(node, AstCode.LdC)) {
+            final Object operand = ((Expression) node).getOperand();
+
+            if (operand instanceof Number) {
+                constant.accept((Number) operand);
+                return true;
+            }
+        }
+
+        return false;
+    }
+
+    public static boolean matchLoad(final Node node, final Predicate<? super Variable> condition) {
+        final StrongBox<Variable> v = new StrongBox<>();
+        return matchLoad(node, v) && condition.test(v.get());
+    }
+
+    public static boolean matchLoad(final Node node, final StrongBox<? super Object> temp, final Predicate<? super Variable> condition) {
+        return matchLoad(node, temp) && condition.test((Variable) temp.get());
+    }
+
     public static Boolean matchBooleanConstant(final Node node) {
         if (match(node, AstCode.LdC)) {
             final Object operand = ((Expression) node).getOperand();
@@ -781,27 +950,27 @@ public final class PatternMatching {
         return null;
     }
 
-    public static boolean matchBooleanConstant(final Node node, final StrongBox<Boolean> value) {
+    public static boolean matchBooleanConstant(final Node node, final Consumer<? super Boolean> value) {
         final Boolean booleanConstant = matchBooleanConstant(node);
 
         if (booleanConstant != null) {
-            value.set(booleanConstant);
+            value.accept(booleanConstant);
             return true;
         }
 
-        value.set(null);
+        value.accept(null);
         return false;
     }
 
-    public static boolean matchCharacterConstant(final Node node, final StrongBox<Character> value) {
+    public static boolean matchCharacterConstant(final Node node, final Consumer<? super Character> value) {
         final Character characterConstant = matchCharacterConstant(node);
 
         if (characterConstant != null) {
-            value.set(characterConstant);
+            value.accept(characterConstant);
             return true;
         }
 
-        value.set(null);
+        value.accept(null);
         return false;
     }
 
@@ -810,11 +979,11 @@ public final class PatternMatching {
                ((Expression) node).getCode().isUnconditionalControlFlow();
     }
 
-    public static boolean matchLock(final List<Node> body, final int position, final StrongBox<LockInfo> result) {
+    public static boolean matchLock(final List<Node> body, final int position, final Consumer<? super LockInfo> result) {
         VerifyArgument.notNull(body, "body");
         VerifyArgument.notNull(result, "result");
 
-        result.set(null);
+        result.accept(null);
 
         int head = position;
 
@@ -842,7 +1011,7 @@ public final class PatternMatching {
                 return false;
             }
 
-            result.set(new LockInfo(leadingLabel, (Expression) body.get(head)));
+            result.accept(new LockInfo(leadingLabel, (Expression) body.get(head)));
             return true;
         }
 
@@ -883,7 +1052,7 @@ public final class PatternMatching {
                     }
                 }
 
-                result.set(
+                result.accept(
                     new LockInfo(
                         leadingLabel,
                         lockInit,
