@@ -18,9 +18,11 @@ package com.strobel.decompiler.languages.java;
 
 import com.strobel.assembler.metadata.MemberReference;
 import com.strobel.assembler.metadata.MethodReference;
+import com.strobel.assembler.metadata.ModuleReference;
 import com.strobel.assembler.metadata.PackageReference;
 import com.strobel.assembler.metadata.ParameterDefinition;
 import com.strobel.assembler.metadata.TypeReference;
+import com.strobel.core.StringUtilities;
 import com.strobel.core.VerifyArgument;
 import com.strobel.decompiler.ITextOutput;
 import com.strobel.decompiler.ast.Variable;
@@ -40,7 +42,7 @@ public class TextOutputFormatter implements IOutputFormatter {
     private boolean firstUsingDeclaration;
     private boolean lastUsingDeclaration;
     private LineNumberMode lineNumberMode;
-    
+
     /**
      * whether or not to emit debug line number comments into the source code
      */
@@ -48,20 +50,26 @@ public class TextOutputFormatter implements IOutputFormatter {
         WITH_DEBUG_LINE_NUMBERS,
         WITHOUT_DEBUG_LINE_NUMBERS,
     }
-    
-    /** when writing out line numbers, keeps track of the most recently used one to avoid redundancy */
+
+    /**
+     * when writing out line numbers, keeps track of the most recently used one to avoid redundancy
+     */
     private int lastObservedLineNumber = OffsetToLineNumberConverter.UNKNOWN_LINE_NUMBER;
-    
-    /** converts from bytecode offset to line number */
+
+    /**
+     * converts from bytecode offset to line number
+     */
     private OffsetToLineNumberConverter offset2LineNumber = OffsetToLineNumberConverter.NOOP_CONVERTER;
-    
-    /** maps original line numbers to decompiler-emitted line numbers and columns */
-    private final List<LineNumberPosition> lineNumberPositions = new ArrayList<LineNumberPosition>(); 
+
+    /**
+     * maps original line numbers to decompiler-emitted line numbers and columns
+     */
+    private final List<LineNumberPosition> lineNumberPositions = new ArrayList<>();
 
     @SuppressWarnings("MismatchedQueryAndUpdateOfCollection")
     private final Stack<TextLocation> startLocations = new Stack<>();
 
-    public TextOutputFormatter(final ITextOutput output, LineNumberMode lineNumberMode) {
+    public TextOutputFormatter(final ITextOutput output, final LineNumberMode lineNumberMode) {
         this.output = VerifyArgument.notNull(output, "output");
         this.lineNumberMode = lineNumberMode;
     }
@@ -80,35 +88,36 @@ public class TextOutputFormatter implements IOutputFormatter {
         }
 
         nodeStack.push(node);
-        
+
         // Build a data structure of Java line numbers.
         int offset = Expression.MYSTERY_OFFSET;
         String prefix = null;
-        if ( node instanceof Expression) {
+        if (node instanceof Expression) {
             offset = ((Expression) node).getOffset();
             prefix = "/*EL:";
-        } else if ( node instanceof Statement) {
+        }
+        else if (node instanceof Statement) {
             offset = ((Statement) node).getOffset();
             prefix = "/*SL:";
         }
-        if ( offset != Expression.MYSTERY_OFFSET) {
+        if (offset != Expression.MYSTERY_OFFSET) {
             // Convert to a line number.
-            int lineNumber = offset2LineNumber.getLineForOffset( offset);
-            if ( lineNumber > lastObservedLineNumber) {
+            final int lineNumber = offset2LineNumber.getLineForOffset(offset);
+            if (lineNumber > lastObservedLineNumber) {
                 // Record a data structure mapping original to actual line numbers.
-                int lineOfComment = output.getRow();
-                int columnOfComment = output.getColumn();
-                LineNumberPosition pos = new LineNumberPosition( lineNumber, lineOfComment, columnOfComment);
-                lineNumberPositions.add( pos);                
+                final int lineOfComment = output.getRow();
+                final int columnOfComment = output.getColumn();
+                final LineNumberPosition pos = new LineNumberPosition(lineNumber, lineOfComment, columnOfComment);
+                lineNumberPositions.add(pos);
                 lastObservedLineNumber = lineNumber;
-                if ( lineNumberMode == LineNumberMode.WITH_DEBUG_LINE_NUMBERS) {
+                if (lineNumberMode == LineNumberMode.WITH_DEBUG_LINE_NUMBERS) {
                     // Emit a comment showing the original line number.
-                    String commentStr = prefix + lineNumber + "*/";
-                    output.writeComment( commentStr);
+                    final String commentStr = prefix + lineNumber + "*/";
+                    output.writeComment(commentStr);
                 }
             }
         }
-        
+
         startLocations.push(new TextLocation(output.getRow(), output.getColumn()));
 
         if (node instanceof EntityDeclaration &&
@@ -151,6 +160,13 @@ public class TextOutputFormatter implements IOutputFormatter {
             return;
         }
 
+        reference = getCurrentModuleReference();
+
+        if (reference != null) {
+            output.writeReference(identifier, reference);
+            return;
+        }
+
         reference = getCurrentTypeReference();
 
         if (reference != null) {
@@ -228,6 +244,50 @@ public class TextOutputFormatter implements IOutputFormatter {
         output.writeTextLiteral(value);
     }
 
+    @Override
+    public void writeTextBlock(final String value) {
+        final int columnStart = output.getColumn();
+        final List<String> lines = StringUtilities.split(value, false, '\n');
+        final boolean endsWithNewline = value.endsWith("\n");
+
+        output.writeTextLiteral("\"\"\"");
+        output.writeLine();
+
+        final int columnEnd = output.getColumn();
+        final int paddingSize = Math.max(0, columnStart - columnEnd);
+        final String padding = StringUtilities.repeat(' ', paddingSize);
+
+        final int n = lines.size();
+
+        for (int i = 0; i < n - 1; i++) {
+            String line = lines.get(i);
+
+            if (line.indexOf('\r') >= 0) {
+                line = line.replace("\r", "\\r");
+            }
+
+            if (line.contains("\"\"\"")) {
+                line = line.replace("\"\"\"", "\\\"\"\"");
+            }
+
+            output.write(padding);
+            output.writeTextLiteral(line);
+            output.writeLine();
+        }
+
+        final String lastLine = lines.get(n - 1);
+
+        output.write(padding);
+        output.writeTextLiteral(lastLine);
+
+        if (endsWithNewline && !StringUtilities.isNullOrWhitespace(lastLine)) {
+            output.writeLine();
+            output.write(padding);
+        }
+
+        output.writeTextLiteral("\"\"\"");
+    }
+
     @Override
     public void space() {
         output.write(' ');
@@ -430,6 +490,16 @@ public class TextOutputFormatter implements IOutputFormatter {
                 parent = parent.getParent();
             }
 
+            if (parent == null) {
+                return null;
+            }
+
+            definition = parent.getUserData(Keys.MODULE_REFERENCE);
+
+            if (definition != null) {
+                return definition;
+            }
+
             definition = parent.getUserData(Keys.TYPE_DEFINITION);
 
             if (definition != null) {
@@ -474,6 +544,25 @@ public class TextOutputFormatter implements IOutputFormatter {
         return null;
     }
 
+    private ModuleReference getCurrentModuleReference() {
+        final AstNode node = nodeStack.peek();
+        final ModuleReference moduleReference = node.getUserData(Keys.MODULE_REFERENCE);
+
+        if (moduleReference != null) {
+            return moduleReference;
+        }
+
+        if (node instanceof Identifier) {
+            final AstNode parent = node.getParent();
+
+            if (parent instanceof ModuleDeclaration) {
+                return parent.getUserData(Keys.MODULE_REFERENCE);
+            }
+        }
+
+        return null;
+    }
+
     private PackageReference getCurrentPackageReference() {
         final AstNode node = nodeStack.peek();
 
@@ -576,13 +665,13 @@ public class TextOutputFormatter implements IOutputFormatter {
     }
 
     @Override
-    public void resetLineNumberOffsets( OffsetToLineNumberConverter offset2LineNumber) {
+    public void resetLineNumberOffsets(final OffsetToLineNumberConverter offset2LineNumber) {
         // Forget what we used to know about the stream of line number offsets and start from
         // scratch.  Also capture the new converter.,
         lastObservedLineNumber = OffsetToLineNumberConverter.UNKNOWN_LINE_NUMBER;
         this.offset2LineNumber = offset2LineNumber;
     }
-    
+
     /**
      * Returns the mapping from original to decompiler-emitted line numbers.
      */
