@@ -19,12 +19,16 @@ package com.strobel.decompiler.languages.java.ast;
 import com.strobel.assembler.ir.attributes.AnnotationDefaultAttribute;
 import com.strobel.assembler.ir.attributes.AttributeNames;
 import com.strobel.assembler.ir.attributes.LineNumberTableAttribute;
+import com.strobel.assembler.ir.attributes.ModuleAttribute;
 import com.strobel.assembler.ir.attributes.SourceAttribute;
 import com.strobel.assembler.metadata.*;
+import com.strobel.assembler.metadata.UnionType;
 import com.strobel.assembler.metadata.annotations.*;
 import com.strobel.core.ArrayUtilities;
+import com.strobel.core.Closeables;
 import com.strobel.core.MutableInteger;
 import com.strobel.core.Predicate;
+import com.strobel.core.SafeCloseable;
 import com.strobel.core.StringUtilities;
 import com.strobel.core.VerifyArgument;
 import com.strobel.decompiler.DecompilerContext;
@@ -37,11 +41,12 @@ import com.strobel.decompiler.languages.java.ast.transforms.IAstTransform;
 import com.strobel.decompiler.languages.java.ast.transforms.TransformationPipeline;
 import com.strobel.util.ContractUtils;
 
-import javax.lang.model.element.Modifier;
 import java.lang.ref.Reference;
 import java.lang.ref.SoftReference;
 import java.util.*;
 
+import static com.strobel.core.CollectionUtilities.first;
+
 public final class AstBuilder {
     private final DecompilerContext _context;
     private final CompilationUnit _compileUnit = new CompilationUnit();
@@ -51,6 +56,7 @@ public final class AstBuilder {
 
     private boolean _decompileMethodBodies = true;
     private boolean _haveTransformationsRun;
+    private int _suppressImportsDepth;
 
     public AstBuilder(final DecompilerContext context) {
         _context = VerifyArgument.notNull(context, "context");
@@ -79,6 +85,23 @@ public final class AstBuilder {
         return _context;
     }
 
+    public final boolean areImportsSuppressed() {
+        return _suppressImportsDepth > 0;
+    }
+
+    public final SafeCloseable suppressImports() {
+        ++_suppressImportsDepth;
+
+        return Closeables.create(
+            new Runnable() {
+                @Override
+                public void run() {
+                    --_suppressImportsDepth;
+                }
+            }
+        );
+    }
+
     public final boolean getDecompileMethodBodies() {
         return _decompileMethodBodies;
     }
@@ -102,7 +125,16 @@ public final class AstBuilder {
     }
 
     public final void addType(final TypeDefinition type) {
-        final TypeDeclaration astType = createType(type);
+        if (type.isModule()) {
+            final ModuleAttribute attribute = SourceAttribute.find(AttributeNames.Module, type.getSourceAttributes());
+
+            if (attribute != null) {
+                _compileUnit.addChild(createModuleNoCache(type, attribute), CompilationUnit.MODULE_ROLE);
+                return;
+            }
+        }
+
+        final EntityDeclaration astType = createType(type);
         final String packageName = type.getPackageName();
 
         if (_compileUnit.getPackage().isNull() && !StringUtilities.isNullOrWhitespace(packageName)) {
@@ -130,6 +162,28 @@ public final class AstBuilder {
         return createTypeNoCache(type);
     }
 
+    protected final ModuleDeclaration createModuleNoCache(final TypeDefinition type, final ModuleAttribute attribute) {
+        VerifyArgument.notNull(type, "type");
+
+        final TypeDefinition oldCurrentType = _context.getCurrentType();
+
+        _context.setCurrentType(type);
+
+        try {
+            final ModuleDeclaration declaration = new ModuleDeclaration();
+
+            declaration.setName(attribute.getModuleName());
+            declaration.putUserData(Keys.TYPE_DEFINITION, type);
+            declaration.putUserData(Keys.MODULE_REFERENCE, new ModuleReference(attribute.getModuleName(), attribute.getVersion()));
+
+            return declaration;
+
+        }
+        finally {
+            _context.setCurrentType(oldCurrentType);
+        }
+    }
+
     protected final TypeDeclaration createTypeNoCache(final TypeDefinition type) {
         VerifyArgument.notNull(type, "type");
 
@@ -170,7 +224,7 @@ public final class AstBuilder {
             declarations.add(d);
 
             if (p.isFinal()) {
-                EntityDeclaration.addModifier(d, Modifier.FINAL);
+                EntityDeclaration.addModifier(d, Flags.Flag.FINAL);
             }
         }
 
@@ -182,6 +236,47 @@ public final class AstBuilder {
             return AstType.NULL;
         }
 
+        if (type instanceof ICompoundType) {
+            final ICompoundType cType = (ICompoundType) type;
+
+            if (options.getIncludeIntersectionTypes()) {
+                final List<TypeReference> ifReferences = cType.getInterfaces();
+
+                final AstType baseType = cType.getBaseType() == BuiltinTypes.Null ? AstType.NULL
+                                                                                  : convertType(cType.getBaseType(), typeIndex, options);
+                final AstType[] ifTypes = new AstType[ifReferences.size()];
+
+                for (int i = 0; i < ifReferences.size(); i++) {
+                    ifTypes[i] = convertType(ifReferences.get(i), typeIndex, options);
+                }
+
+                final IntersectionType isType = new IntersectionType(baseType, ifTypes);
+                isType.putUserData(Keys.TYPE_REFERENCE, type);
+                return isType;
+            }
+
+            return convertType(cType.getBaseType(), typeIndex, options).makeArrayType();
+        }
+
+        if (type instanceof UnionType) {
+            final UnionType uType = (UnionType) type;
+
+            if (options.getIncludeUnionTypes()) {
+                final List<TypeReference> alternatives = uType.getAlternatives();
+                final AstType[] astAlternatives = new AstType[alternatives.size()];
+
+                for (int i = 0; i < alternatives.size(); i++) {
+                    astAlternatives[i] = convertType(alternatives.get(i), typeIndex, options);
+                }
+
+                final com.strobel.decompiler.languages.java.ast.UnionType isType = new com.strobel.decompiler.languages.java.ast.UnionType(astAlternatives);
+                isType.putUserData(Keys.TYPE_REFERENCE, type);
+                return isType;
+            }
+
+            return convertType(first(uType.getAlternatives()), typeIndex, options).makeArrayType();
+        }
+
         if (type.isArray()) {
             return convertType(type.getElementType(), typeIndex.increment(), options).makeArrayType();
         }
@@ -300,17 +395,17 @@ public final class AstBuilder {
         else {
             final TypeReference typeToImport;
 
-            String unqualifiedName;
+            StringBuilder unqualifiedName;
 
             if (packageDeclaration != null &&
                 StringUtilities.equals(packageDeclaration.getName(), nameSource.getPackageName())) {
 
-                unqualifiedName = nameSource.getSimpleName();
-                name = unqualifiedName;
+                unqualifiedName = new StringBuilder(nameSource.getSimpleName());
+                name = unqualifiedName.toString();
             }
 
             if (nameSource.isNested()) {
-                unqualifiedName = nameSource.getSimpleName();
+                unqualifiedName = new StringBuilder(nameSource.getSimpleName());
 
                 TypeReference current = nameSource;
 
@@ -321,18 +416,18 @@ public final class AstBuilder {
                         break;
                     }
 
-                    unqualifiedName = current.getSimpleName() + "." + unqualifiedName;
+                    unqualifiedName.insert(0, current.getSimpleName() + ".");
                 }
 
-                name = unqualifiedName;
+                name = unqualifiedName.toString();
                 typeToImport = current;
             }
             else {
                 typeToImport = nameSource;
-                unqualifiedName = nameSource.getSimpleName();
+                unqualifiedName = new StringBuilder(nameSource.getSimpleName());
             }
 
-            if (options.getAddImports() && !_typeDeclarations.containsKey(typeToImport.getInternalName())) {
+            if (options.getAddImports() && !areImportsSuppressed() && !_typeDeclarations.containsKey(typeToImport.getInternalName())) {
                 String importedName = _unqualifiedTypeNames.get(typeToImport.getSimpleName());
 
                 if (importedName == null) {
@@ -357,7 +452,7 @@ public final class AstBuilder {
 
                 if (name == null) {
                     if (importedName.equals(typeToImport.getFullName())) {
-                        name = unqualifiedName;
+                        name = unqualifiedName.toString();
                     }
                     else {
                         final String packageName = nameSource.getPackageName();
@@ -371,6 +466,10 @@ public final class AstBuilder {
             }
         }
 
+        if (name == null) {
+            name = nameSource.getSimpleName();
+        }
+
         final SimpleType astType = new SimpleType(name);
 
         astType.putUserData(Keys.TYPE_REFERENCE, type);
@@ -444,16 +543,19 @@ public final class AstBuilder {
 
         long flags = type.getFlags();
 
-        if (type.isInterface() || type.isEnum()) {
+        if (type.isInterface()) {
+            flags &= Flags.AccessFlags | Flags.ExtendedInterfaceFlags;
+        }
+        else if (type.isEnum()) {
             flags &= Flags.AccessFlags;
         }
         else {
-            flags &= (Flags.AccessFlags | Flags.ClassFlags | Flags.STATIC | Flags.FINAL);
+            flags &= (Flags.AccessFlags | Flags.ExtendedClassFlags | Flags.STATIC | Flags.FINAL);
         }
 
         EntityDeclaration.setModifiers(
             astType,
-            Flags.asModifierSet(scrubAccessModifiers(flags))
+            Flags.asFlagSet(scrubAccessModifiers(flags))
         );
 
         astType.setName(type.getSimpleName());
@@ -486,12 +588,16 @@ public final class AstBuilder {
         }
 
         for (final TypeReference interfaceType : type.getExplicitInterfaces()) {
-            if (type.isAnnotation() && "java/lang/annotations/Annotation".equals(interfaceType.getInternalName())) {
+            if (type.isAnnotation() && CommonTypeReferences.Annotation.isEquivalentTo(interfaceType)) {
                 continue;
             }
             astType.addChild(convertType(interfaceType), Roles.IMPLEMENTED_INTERFACE);
         }
 
+        for (final TypeReference permittedSubclass : type.getPermittedSubclasses()) {
+            astType.addChild(convertType(permittedSubclass), Roles.PERMITTED_SUBCLASSES);
+        }
+
         for (final CustomAnnotation annotation : type.getAnnotations()) {
             astType.getAnnotations().add(createAnnotation(annotation));
         }
@@ -605,7 +711,7 @@ public final class AstBuilder {
 
         EntityDeclaration.setModifiers(
             astField,
-            Flags.asModifierSet(scrubAccessModifiers(field.getFlags() & Flags.VarFlags))
+            Flags.asFlagSet(scrubAccessModifiers(field.getFlags() & Flags.VarFlags))
         );
 
         if (field.hasConstantValue()) {
@@ -621,19 +727,19 @@ public final class AstBuilder {
         return astField;
     }
 
-    private MethodDeclaration createMethod(final MethodDefinition method) {
+    public final MethodDeclaration createMethod(final MethodDefinition method) {
         final MethodDeclaration astMethod = new MethodDeclaration();
 
-        final Set<Modifier> modifiers;
+        final Set<Flags.Flag> modifiers;
 
         if (method.isTypeInitializer()) {
-            modifiers = Collections.singleton(Modifier.STATIC);
+            modifiers = Collections.singleton(Flags.Flag.STATIC);
         }
         else if (method.getDeclaringType().isInterface()) {
             modifiers = Collections.emptySet();
         }
         else {
-            modifiers = Flags.asModifierSet(scrubAccessModifiers(method.getFlags() & Flags.MethodFlags));
+            modifiers = Flags.asFlagSet(scrubAccessModifiers(method.getFlags() & Flags.MethodFlags));
         }
 
         EntityDeclaration.setModifiers(astMethod, modifiers);
@@ -684,15 +790,20 @@ public final class AstBuilder {
 
         EntityDeclaration.setModifiers(
             astMethod,
-            Flags.asModifierSet(scrubAccessModifiers(method.getFlags() & Flags.ConstructorFlags))
+            Flags.asFlagSet(scrubAccessModifiers(method.getFlags() & Flags.ConstructorFlags))
         );
 
         astMethod.setName(method.getDeclaringType().getName());
         astMethod.getParameters().addAll(createParameters(method.getParameters()));
+        astMethod.getTypeParameters().addAll(createTypeParameters(method.getGenericParameters()));
         astMethod.setBody(createMethodBody(method, astMethod.getParameters()));
         astMethod.putUserData(Keys.METHOD_DEFINITION, method);
         astMethod.putUserData(Keys.MEMBER_REFERENCE, method);
 
+        for (final CustomAnnotation annotation : method.getAnnotations()) {
+            astMethod.getAnnotations().add(createAnnotation(annotation));
+        }
+
         for (final TypeReference thrownType : method.getThrownTypes()) {
             astMethod.addChild(convertType(thrownType), Roles.THROWN_TYPE);
         }
