@@ -49,7 +49,7 @@ public class DefiniteAssignmentAnalysis {
     private final List<IdentifierExpression> unassignedVariableUsesView = Collections.unmodifiableList(unassignedVariableUses);
     private final ArrayDeque<DefiniteAssignmentNode> nodesWithModifiedInput = new ArrayDeque<>();
 
-    private Function<AstNode, ResolveResult> resolver;
+    private final Function<AstNode, ResolveResult> resolver;
 
     private String variableName;
 
@@ -79,6 +79,18 @@ public class DefiniteAssignmentAnalysis {
 
             node.setIndex(i);
 
+            if (node.getType() == ControlFlowNodeType.StartNode ||
+                node.getType() == ControlFlowNodeType.BetweenStatements) {
+
+                //
+                // Anonymous methods have separate control flow graphs, but we also need to analyze those.
+                // Iterate backwards so that anonymous methods are inserted in the correct order.
+                //
+                for (AstNode child = node.getNextStatement().getLastChild(); child != null; child = child.getPreviousSibling()) {
+                    insertAnonymousMethods(i + 1, child, builder);
+                }
+            }
+
             if (node.getType() == ControlFlowNodeType.StartNode ||
                 node.getType() == ControlFlowNodeType.BetweenStatements) {
 
@@ -100,6 +112,42 @@ public class DefiniteAssignmentAnalysis {
         this.analyzedRangeEnd = allNodes.size() - 1;
     }
 
+    private void insertAnonymousMethods(
+        final int insertPosition,
+        final AstNode node,
+        final ControlFlowGraphBuilder builder) {
+
+        //
+        // Ignore any statements, as those have their own ControlFlowNode and get handled separately.
+        //
+        if (node instanceof Statement) {
+            return;
+        }
+
+        if (node instanceof LambdaExpression){
+            final LambdaExpression lambda = (LambdaExpression) node;
+
+            if (lambda.getBody() instanceof Statement) {
+                @SuppressWarnings({ "unchecked", "rawtypes" })
+                final List<? extends DefiniteAssignmentNode> nodes = (List) builder.buildControlFlowGraph(
+                    (Statement) lambda.getBody(),
+                    resolver
+                );
+
+                allNodes.addAll(insertPosition, nodes);
+
+                return;
+            }
+        }
+
+        //
+        // Descend into child expressions.  Iterate backwards so that anonymous methods
+        // are inserted in the correct order.
+        //
+        for (AstNode child = node.getLastChild(); child != null; child = child.getPreviousSibling()) {
+            insertAnonymousMethods(insertPosition, child, builder);
+        }
+    }
     public List<IdentifierExpression> getUnassignedVariableUses() {
         return unassignedVariableUsesView;
     }
@@ -251,6 +299,7 @@ public class DefiniteAssignmentAnalysis {
                 //
             }
 
+            //noinspection fallthrough
             case LoopCondition: {
                 if (node.getNextStatement() instanceof ForEachStatement) {
                     final ForEachStatement forEach = (ForEachStatement) node.getNextStatement();
@@ -477,7 +526,13 @@ public class DefiniteAssignmentAnalysis {
 
         @Override
         public DefiniteAssignmentStatus visitTryCatchStatement(final TryCatchStatement node, final DefiniteAssignmentStatus data) {
-            return data;
+            DefiniteAssignmentStatus status = data;
+
+            for (final VariableDeclarationStatement resource : node.getDeclaredResources()) {
+                status = mergeStatus(status, resource.acceptVisitor(this, data));
+            }
+
+            return status;
         }
 
         @Override
@@ -500,6 +555,18 @@ public class DefiniteAssignmentAnalysis {
             }
         }
 
+        @Override
+        public DefiniteAssignmentStatus visitLambdaExpression(final LambdaExpression node, final DefiniteAssignmentStatus data) {
+            final AstNode body = node.getBody();
+            if (body instanceof Statement) {
+                changeNodeStatus(beginNodeMap.get(body), data);
+            }
+            else {
+                body.acceptVisitor(this, data);
+            }
+            return data;
+        }
+
         final DefiniteAssignmentStatus handleAssignment(final Expression left, final Expression right, final DefiniteAssignmentStatus initialStatus) {
             if (left instanceof IdentifierExpression) {
                 final IdentifierExpression identifier = (IdentifierExpression) left;
@@ -725,7 +792,7 @@ public class DefiniteAssignmentAnalysis {
 
     // <editor-fold defaultstate="collapsed" desc="DefiniteAssignmentNode Class">
 
-    final class DefiniteAssignmentNode extends ControlFlowNode {
+    final static class DefiniteAssignmentNode extends ControlFlowNode {
         private int _index;
         private DefiniteAssignmentStatus _nodeStatus;
 
@@ -763,7 +830,7 @@ public class DefiniteAssignmentAnalysis {
 
     // <editor-fold defaultstate="collapsed" desc="DerivedControlFlowGraphBuilder Class">
 
-    final class DerivedControlFlowGraphBuilder extends ControlFlowGraphBuilder {
+    final static class DerivedControlFlowGraphBuilder extends ControlFlowGraphBuilder {
         @Override
         protected ControlFlowNode createNode(
             final Statement previousStatement,
