@@ -40,9 +40,11 @@ public class EclipseEnumSwitchRewriterTransform implements IAstTransform {
 
     @Override
     public void run(final AstNode compilationUnit) {
-        final Visitor visitor = new Visitor(_context);
-        compilationUnit.acceptVisitor(visitor, null);
-        visitor.rewrite();
+        if (_context.isSupported(LanguageFeature.ENUM_CLASSES)) {
+            final Visitor visitor = new Visitor(_context);
+            compilationUnit.acceptVisitor(visitor, null);
+            visitor.rewrite();
+        }
     }
 
     private final static class Visitor extends ContextTrackingVisitor<Void> {
@@ -134,7 +136,7 @@ public class EclipseEnumSwitchRewriterTransform implements IAstTransform {
         }
 
         @Override
-        public Void visitMethodDeclaration(final MethodDeclaration node, final Void p) {
+        protected Void visitMethodDeclarationOverride(final MethodDeclaration node, final Void p) {
             final MethodDefinition methodDefinition = node.getUserData(Keys.METHOD_DEFINITION);
 
             if (isSwitchMapMethod(methodDefinition)) {
@@ -169,7 +171,7 @@ public class EclipseEnumSwitchRewriterTransform implements IAstTransform {
                 }
             }
 
-            return super.visitMethodDeclaration(node, p);
+            return super.visitMethodDeclarationOverride(node, p);
         }
 
         private void rewrite() {
@@ -404,7 +406,7 @@ public class EclipseEnumSwitchRewriterTransform implements IAstTransform {
                             )
                         )
                     ),
-                    new TypedPrimitiveValueNode("tableValue", Integer.class).toExpression()
+                    new TypedLiteralNode("tableValue", Integer.class).toExpression()
                 )
             );
 
