@@ -17,11 +17,12 @@
 package com.strobel.decompiler.languages.java.ast.transforms;
 
 import com.strobel.assembler.metadata.*;
+import com.strobel.core.SafeCloseable;
+import com.strobel.core.StringUtilities;
 import com.strobel.decompiler.DecompilerContext;
 import com.strobel.decompiler.languages.java.ast.*;
 import com.strobel.decompiler.patterns.*;
 
-import javax.lang.model.element.Modifier;
 import java.util.ArrayList;
 import java.util.HashMap;
 import java.util.HashSet;
@@ -83,20 +84,22 @@ public class EliminateSyntheticAccessorsTransform extends ContextTrackingVisitor
                 final MethodReference method = (MethodReference) reference;
                 final TypeReference declaringType = method.getDeclaringType();
 
-                if (!MetadataResolver.areEquivalent(context.getCurrentType(), declaringType) &&
+                if (StringUtilities.equals(declaringType.getPackageName(), context.getCurrentType().getPackageName()) &&
+                    !MetadataResolver.areEquivalent(context.getCurrentType(), declaringType) &&
                     !MetadataHelper.isEnclosedBy(context.getCurrentType(), declaringType) &&
                     !_visitedTypes.contains(declaringType.getInternalName())) {
 
-                    final MethodDefinition resolvedMethod = method.resolve();
+                    final MethodDefinition md = method.resolve();
 
-                    if (resolvedMethod != null &&
-                        resolvedMethod.isSynthetic()) {
+                    if (md != null && md.isSynthetic() && !md.isBridgeMethod() && md.isPackagePrivate()) {
                         final AstBuilder astBuilder = context.getUserData(Keys.AST_BUILDER);
 
                         if (astBuilder != null) {
-                            final TypeDeclaration ownerTypeDeclaration = astBuilder.createType(resolvedMethod.getDeclaringType());
+                            try (final SafeCloseable importSuppression = astBuilder.suppressImports()) {
+                                final TypeDeclaration ownerTypeDeclaration = astBuilder.createType(md.getDeclaringType());
 
-                            ownerTypeDeclaration.acceptVisitor(new PhaseOneVisitor(), data);
+                                ownerTypeDeclaration.acceptVisitor(new PhaseOneVisitor(), data);
+                            }
                         }
                     }
                 }
@@ -278,11 +281,11 @@ public class EliminateSyntheticAccessorsTransform extends ContextTrackingVisitor
         final MethodDeclaration setAccessor = new MethodDeclaration();
 
         getAccessor.setName(Pattern.ANY_STRING);
-        getAccessor.getModifiers().add(new JavaModifierToken(Modifier.STATIC));
+        getAccessor.getModifiers().add(new JavaModifierToken(Flags.Flag.STATIC));
         getAccessor.setReturnType(new AnyNode("returnType").toType());
 
         setAccessor.setName(Pattern.ANY_STRING);
-        setAccessor.getModifiers().add(new JavaModifierToken(Modifier.STATIC));
+        setAccessor.getModifiers().add(new JavaModifierToken(Flags.Flag.STATIC));
         setAccessor.setReturnType(new AnyNode("returnType").toType());
 
         final ParameterDeclaration getParameter = new ParameterDeclaration(
@@ -300,7 +303,7 @@ public class EliminateSyntheticAccessorsTransform extends ContextTrackingVisitor
 
         final ParameterDeclaration setParameter2 = new ParameterDeclaration(
             Pattern.ANY_STRING,
-            new BackReference("returnType").toType()
+            new AnyNode().toType()
         );
 
         setParameter1.setAnyModifiers(true);
@@ -343,7 +346,13 @@ public class EliminateSyntheticAccessorsTransform extends ContextTrackingVisitor
                                 FieldReference.class
                             ).toExpression(),
                             AssignmentOperatorType.ANY,
-                            new ParameterReferenceNode(1, "value").toExpression()
+                            new Choice(
+                                new ParameterReferenceNode(1, "value"),
+                                new CastExpression(
+                                    new BackReference("returnType").toType(),
+                                    new ParameterReferenceNode(1, "value").toExpression()
+                                )
+                            ).toExpression()
                         )
                     ),
                     new ReturnStatement(Expression.MYSTERY_OFFSET, new BackReference("value").toExpression())
@@ -361,7 +370,13 @@ public class EliminateSyntheticAccessorsTransform extends ContextTrackingVisitor
                                 FieldReference.class
                             ).toExpression(),
                             AssignmentOperatorType.ANY,
-                            new ParameterReferenceNode(1, "value").toExpression()
+                            new Choice(
+                                new ParameterReferenceNode(1, "value"),
+                                new CastExpression(
+                                    new BackReference("returnType").toType(),
+                                    new ParameterReferenceNode(1, "value").toExpression()
+                                )
+                            ).toExpression()
                         )
                     )
                 )
@@ -374,7 +389,7 @@ public class EliminateSyntheticAccessorsTransform extends ContextTrackingVisitor
             new AnyNode("value").toExpression()
         );
 
-        tempVariable.addModifier(Modifier.FINAL);
+        tempVariable.addModifier(Flags.Flag.FINAL);
 
         altSetAccessor.setBody(
             new BlockStatement(
@@ -500,7 +515,7 @@ public class EliminateSyntheticAccessorsTransform extends ContextTrackingVisitor
         }
 
         @Override
-        public Void visitTypeDeclaration(final TypeDeclaration node, final Void p) {
+        protected Void visitTypeDeclarationOverride(final TypeDeclaration node, final Void p) {
             final TypeDefinition type = node.getUserData(Keys.TYPE_DEFINITION);
 
             if (type != null) {
@@ -509,11 +524,11 @@ public class EliminateSyntheticAccessorsTransform extends ContextTrackingVisitor
                 }
             }
 
-            return super.visitTypeDeclaration(node, p);
+            return super.visitTypeDeclarationOverride(node, p);
         }
 
         @Override
-        public Void visitMethodDeclaration(final MethodDeclaration node, final Void p) {
+        protected Void visitMethodDeclarationOverride(final MethodDeclaration node, final Void p) {
             final MethodDefinition method = node.getUserData(Keys.METHOD_DEFINITION);
 
             if (method != null) {
@@ -524,7 +539,7 @@ public class EliminateSyntheticAccessorsTransform extends ContextTrackingVisitor
                 }
             }
 
-            return super.visitMethodDeclaration(node, p);
+            return super.visitMethodDeclarationOverride(node, p);
         }
 
         private boolean tryMatchAccessor(final MethodDeclaration node) {
