@@ -32,7 +32,9 @@ import com.strobel.decompiler.patterns.INode;
 import com.strobel.decompiler.semantics.ResolveResult;
 import com.strobel.functions.Function;
 
-import static com.strobel.core.CollectionUtilities.firstOrDefault;
+import java.util.ArrayDeque;
+
+import static com.strobel.core.CollectionUtilities.*;
 
 public class InsertNecessaryConversionsTransform extends ContextTrackingVisitor<Void> {
     private final static ConvertTypeOptions NO_IMPORT_OPTIONS;
@@ -48,10 +50,12 @@ public class InsertNecessaryConversionsTransform extends ContextTrackingVisitor<
     }
 
     private final JavaResolver _resolver;
+    private final ArrayDeque<INode> _stack;
 
     public InsertNecessaryConversionsTransform(final DecompilerContext context) {
         super(context);
         _resolver = new JavaResolver(context);
+        _stack = new ArrayDeque<>();
     }
 
     @Override
@@ -103,34 +107,57 @@ public class InsertNecessaryConversionsTransform extends ContextTrackingVisitor<
             return null;
         }
 
-        final ResolveResult valueResult = _resolver.apply(node.getTarget());
+        final Expression target = node.getTarget();
+        final ResolveResult valueResult = _resolver.apply(target);
 
         TypeReference declaringType = member.getDeclaringType();
 
         if (valueResult != null &&
             valueResult.getType() != null) {
 
+            if (target instanceof LambdaExpression || target instanceof MethodGroupExpression) {
+                final TypeReference finalDeclaringType = adjustDeclaringType(valueResult, declaringType);
+
+                target.replaceWith(
+                    new Function<AstNode, AstNode>() {
+                        @Override
+                        public AstNode apply(final AstNode input) {
+                            return new CastExpression(astBuilder.convertType(finalDeclaringType), target);
+                        }
+                    }
+                );
+
+                recurse(node);
+
+                return null;
+            }
+
             if (MetadataHelper.isAssignableFrom(declaringType, valueResult.getType())) {
                 return null;
             }
 
-            if (valueResult.getType().isGenericType() &&
-                (declaringType.isGenericType() ||
-                 MetadataHelper.isRawType(declaringType))) {
-
-                final TypeReference asSuper = MetadataHelper.asSuper(declaringType, valueResult.getType());
-
-                if (asSuper != null) {
-                    declaringType = asSuper;
-                }
-            }
+            declaringType = adjustDeclaringType(valueResult, declaringType);
         }
 
-        addCastForAssignment(astBuilder.convertType(declaringType, NO_IMPORT_OPTIONS), node.getTarget());
+        addCastForAssignment(astBuilder.convertType(declaringType, NO_IMPORT_OPTIONS), target);
 
         return null;
     }
 
+    private static TypeReference adjustDeclaringType(final ResolveResult valueResult, final TypeReference declaringType) {
+        if (valueResult.getType().isGenericType() &&
+            (declaringType.isGenericType() ||
+             MetadataHelper.isRawType(declaringType))) {
+
+            final TypeReference asSuper = MetadataHelper.asSuper(declaringType, valueResult.getType());
+
+            if (asSuper != null) {
+                return asSuper;
+            }
+        }
+        return declaringType;
+    }
+
     @Override
     public Void visitAssignmentExpression(final AssignmentExpression node, final Void data) {
         super.visitAssignmentExpression(node, data);
@@ -201,6 +228,23 @@ public class InsertNecessaryConversionsTransform extends ContextTrackingVisitor<
         return null;
     }
 
+    @Override
+    public Void visitArrayInitializerExpression(final ArrayInitializerExpression node, final Void data) {
+        super.visitArrayInitializerExpression(node, data);
+
+        final ArrayCreationExpression creation = firstOrDefault(ofType(node.getAncestors(), ArrayCreationExpression.class));
+
+        if (creation == null || !creation.getAdditionalArraySpecifiers().hasSingleElement()) {
+            return null;
+        }
+
+        for (final Expression element : node.getElements()) {
+            addCastForAssignment(creation.getType(), element);
+        }
+
+        return null;
+    }
+
     private boolean addCastForAssignment(final AstNode left, final Expression right) {
         final ResolveResult targetResult = _resolver.apply(left);
 
@@ -531,12 +575,19 @@ public class InsertNecessaryConversionsTransform extends ContextTrackingVisitor<
 
     private void recurse(final AstNode replacement) {
         final AstNode parent = replacement.getParent();
+        final AstNode toRecurse = parent != null ? parent : replacement;
 
-        if (parent != null) {
-            parent.acceptVisitor(this, null);
+        if (_stack.contains(toRecurse)) {
+            return;
         }
-        else {
-            replacement.acceptVisitor(this, null);
+
+        _stack.addFirst(toRecurse);
+
+        try {
+            toRecurse.acceptVisitor(this, null);
+        }
+        finally {
+            _stack.removeFirst();
         }
     }
 }
