@@ -23,6 +23,7 @@ import java.io.IOException;
 import java.io.OutputStreamWriter;
 import java.io.Writer;
 import java.nio.charset.Charset;
+import java.nio.charset.StandardCharsets;
 import java.text.DateFormat;
 import java.text.SimpleDateFormat;
 import java.util.Date;
@@ -64,7 +65,7 @@ public class DecompilerDriver {
         final boolean decompileJar = !StringUtilities.isNullOrWhitespace(jarFile);
 
         if (options.getPrintVersion()) {
-            JCommander.getConsole().println(Procyon.version());
+            JCommander.newBuilder().build().getConsole().println(Procyon.version());
             if (options.getPrintUsage()) {
                 jCommander.usage();
             }
@@ -93,7 +94,10 @@ public class DecompilerDriver {
         settings.setMergeVariables(options.getMergeVariables());
         settings.setShowDebugLineNumbers(options.getShowDebugLineNumbers());
         settings.setSimplifyMemberReferences(options.getSimplifyMemberReferences());
+        settings.setForceFullyQualifiedReferences(options.getForceFullyQualifiedReferences());
         settings.setDisableForEachTransforms(options.getDisableForEachTransforms());
+        settings.setForcedCompilerTarget(options.getCompilerTargetOverride());
+        settings.setTextBlockLineMinimum(options.getTextBlockLineMinimum());
         settings.setTypeLoader(new InputTypeLoader());
 
         if (!options.getSuppressBanner()) {
@@ -227,13 +231,7 @@ public class DecompilerDriver {
         final ITypeLoader oldTypeLoader = settings.getTypeLoader();
 
         settings.setShowSyntheticMembers(false);
-
-        settings.setTypeLoader(
-            new CompositeTypeLoader(
-                new JarTypeLoader(jar),
-                settings.getTypeLoader()
-            )
-        );
+        settings.setTypeLoader(new CompositeTypeLoader(new JarTypeLoader(jar), oldTypeLoader));
 
         try {
             MetadataSystem metadataSystem = new NoRetryMetadataSystem(settings.getTypeLoader());
@@ -371,7 +369,7 @@ public class DecompilerDriver {
         if (StringUtilities.isNullOrWhitespace(outputDirectory)) {
             return new OutputStreamWriter(
                 System.out,
-                settings.isUnicodeOutputEnabled() ? Charset.forName("UTF-8")
+                settings.isUnicodeOutputEnabled() ? StandardCharsets.UTF_8
                                                   : Charset.defaultCharset()
             );
         }
@@ -422,7 +420,7 @@ final class FileOutputWriter extends OutputStreamWriter {
     FileOutputWriter(final File file, final DecompilerSettings settings) throws IOException {
         super(
             new FileOutputStream(file),
-            settings.isUnicodeOutputEnabled() ? Charset.forName("UTF-8")
+            settings.isUnicodeOutputEnabled() ? StandardCharsets.UTF_8
                                               : Charset.defaultCharset()
         );
         this.file = file;
@@ -464,10 +462,6 @@ final class NoRetryMetadataSystem extends MetadataSystem {
     NoRetryMetadataSystem() {
     }
 
-    NoRetryMetadataSystem(final String classPath) {
-        super(classPath);
-    }
-
     NoRetryMetadataSystem(final ITypeLoader typeLoader) {
         super(typeLoader);
     }
