@@ -16,6 +16,8 @@
 
 package com.strobel.expressions;
 
+import com.strobel.core.CollectionUtilities;
+import com.strobel.core.StringUtilities;
 import com.strobel.core.VerifyArgument;
 import com.strobel.reflection.FieldInfo;
 import com.strobel.reflection.MemberInfo;
@@ -23,6 +25,8 @@ import com.strobel.reflection.MemberType;
 import com.strobel.reflection.MethodBase;
 import com.strobel.reflection.Type;
 
+import java.util.Arrays;
+
 import static java.lang.String.format;
 
 /**
@@ -69,15 +73,11 @@ final class Error {
     }
 
     public static IllegalStateException argumentTypesMustMatch() {
-        return new IllegalStateException(
-            format("Argument types must match.")
-        );
+        return new IllegalStateException("Argument types must match.");
     }
 
     public static IllegalStateException argumentCannotBeOfTypeVoid() {
-        return new IllegalStateException(
-            format("Argument cannot be of type 'void'.")
-        );
+        return new IllegalStateException("Argument cannot be of type 'void'.");
     }
 
     public static IllegalStateException expressionMustBeWriteable(final String parameterName) {
@@ -217,6 +217,14 @@ final class Error {
         return new IllegalStateException("Incorrect number of arguments supplied for constructor call.");
     }
 
+    public static IllegalStateException couldNotResolveConstructor(final Type<?> type, final Type<?>... parameterTypes) {
+        return new IllegalStateException(format("Could not resolve '%s' constructor for argument types [%s].",
+                                                type.getName(),
+                                                StringUtilities.join(", ",
+                                                                     CollectionUtilities.select(Arrays.asList(parameterTypes),
+                                                                                                Functions.typeName()))));
+    }
+
     public static IllegalStateException incorrectNumberOfLambdaArguments() {
         return new IllegalStateException("Incorrect number of arguments supplied for lambda invocation.");
     }
@@ -241,6 +249,12 @@ final class Error {
         );
     }
 
+    public static IllegalArgumentException unboxNotDefined(final Type<?> boxedType, final Type<?> unboxedType) {
+        return new IllegalArgumentException(
+            format("Could not resolve an unboxing method from %s to %s.", boxedType.getName(), unboxedType.getName())
+        );
+    }
+
     public static IllegalStateException invalidBoxType() {
         return new IllegalStateException(
             "Can only unbox from a standard boxed type or java.lang.Object to a primitive type."
@@ -367,6 +381,16 @@ final class Error {
         );
     }
 
+    public static IllegalStateException targetInvalidForStaticFieldAccess(final FieldInfo field) {
+        return new IllegalStateException(
+            format(
+                "A target expression cannot be used to access static field '%s.%s'.",
+                field.getDeclaringType().getBriefDescription(),
+                field.getName()
+            )
+        );
+    }
+
     public static IllegalStateException targetInvalidForStaticMethodCall(final MethodBase method) {
         return new IllegalStateException(
             format(
