@@ -50,7 +50,7 @@ public class CodeGenerator {
     private final static int MIN_BYTE = 0x00;
     private final static int MAX_BYTE = 0xFF;
 
-    private CodeStream _codeStream;
+    private final CodeStream _codeStream;
 
     private int[] _labelList;
     private int _labelCount;
@@ -99,6 +99,7 @@ public class CodeGenerator {
 
     // <editor-fold defaultstate="collapsed" desc="Exceptions">
 
+    @SuppressWarnings("UnusedReturnValue")
     public Label beginExceptionBlock() {
         if (_exceptions == null) {
             _exceptions = new __ExceptionInfo[DefaultExceptionArraySize];
@@ -346,6 +347,22 @@ public class CodeGenerator {
         emit(OpCode.DUP2_X2);
     }
 
+    public void dup(final Type<?> type) {
+        switch (type.getKind()) {
+            case LONG:
+            case DOUBLE:
+                emit(OpCode.DUP2);
+                break;
+
+            case VOID:
+                break;
+
+            default:
+                emit(OpCode.DUP);
+                break;
+        }
+    }
+
     public void pop() {
         emit(OpCode.POP);
     }
@@ -450,15 +467,34 @@ public class CodeGenerator {
         int stackChange = 0;
 
         if (constructor instanceof ConstructorBuilder) {
-            stackChange -= ((ConstructorBuilder)constructor).getParameterTypes().size();
+            for (final Type<?> type : ((ConstructorBuilder) constructor).getParameterTypes()) {
+                stackChange -= stackSize(type);
+            }
         }
         else {
+            for (final ParameterInfo p : constructor.getParameters()) {
+                stackChange -= stackSize(p.getParameterType());
+            }
             stackChange -= constructor.getParameters().size();
         }
 
         updateStackSize(opCode, stackChange);
     }
 
+    private static int stackSize(final Type<?> type) {
+        final TypeKind kind = type.getKind();
+
+        if (kind == TypeKind.LONG || kind == TypeKind.DOUBLE) {
+            return 2;
+        }
+
+        if (kind == TypeKind.VOID) {
+            return 0;
+        }
+
+        return 1;
+    }
+
     public void emit(final OpCode opCode, final MethodInfo method) {
         VerifyArgument.notNull(method, "method");
 
@@ -474,6 +510,7 @@ public class CodeGenerator {
 
         final TypeList parameterTypes;
 
+        int stackChange = 0;
         int formalParametersSize = 0;
 
         if (method instanceof MethodBuilder) {
@@ -484,10 +521,10 @@ public class CodeGenerator {
         }
 
         for (final Type<?> parameterType : parameterTypes) {
-            ++formalParametersSize;
-            if (parameterType == PrimitiveTypes.Long || parameterType == PrimitiveTypes.Double) {
-                ++formalParametersSize;
-            }
+            final int size = stackSize(parameterType);
+
+            stackChange -= size;
+            formalParametersSize += size;
         }
 
         if (opCode == OpCode.INVOKEINTERFACE) {
@@ -499,16 +536,7 @@ public class CodeGenerator {
 
         registerCheckedExceptions(method);
 
-        int stackChange = -formalParametersSize;
-
-        final Type<?> returnType = method.getReturnType();
-
-        if (returnType != PrimitiveTypes.Void) {
-            ++stackChange;
-            if (returnType == PrimitiveTypes.Long || returnType == PrimitiveTypes.Double) {
-                ++stackChange;
-            }
-        }
+        stackChange += stackSize(method.getReturnType());
 
         updateStackSize(opCode, stackChange);
     }
@@ -524,7 +552,18 @@ public class CodeGenerator {
 
         final int fieldToken = methodBuilder.getDeclaringType().getFieldToken(field);
 
+
+        int stackChange = 0;
+
+        if (opCode == OpCode.PUTFIELD || opCode == OpCode.PUTSTATIC) {
+            stackChange -= stackSize(field.getFieldType());
+        }
+        else {
+            stackChange += stackSize(field.getFieldType());
+        }
+
         emit(opCode, (short)fieldToken);
+        updateStackSize(opCode, stackChange);
     }
 
     public void emit(final OpCode opCode, final Label label) {
@@ -1167,7 +1206,7 @@ public class CodeGenerator {
             }
 
             final TypeList parameterTypes = methodBuilder.getParameterTypes();
-            
+
             for (int i = 0, n = parameterTypes.size(); i < localIndex && i < n; i++) {
                 final TypeKind kind = parameterTypes.get(i).getKind();
                 if (kind == TypeKind.LONG || kind == TypeKind.DOUBLE) {
@@ -1178,7 +1217,7 @@ public class CodeGenerator {
 
         return index;
     }
-    
+
     final int translateLocal(final int localIndex) {
         int index = 0;
 
@@ -1933,6 +1972,7 @@ public class CodeGenerator {
 
             case BYTE: {
                 switch (sourceKind) {
+                    case BOOLEAN:
                     case CHAR:
                     case SHORT:
                     case INT:
@@ -1954,10 +1994,12 @@ public class CodeGenerator {
                         emit(OpCode.I2B);
                         return;
                 }
+                break;
             }
 
             case SHORT: {
                 switch (sourceKind) {
+                    case BOOLEAN:
                     case BYTE:
                     case CHAR:
                     case INT:
@@ -1979,10 +2021,12 @@ public class CodeGenerator {
                         emit(OpCode.I2S);
                         return;
                 }
+                break;
             }
 
             case INT: {
                 switch (sourceKind) {
+                    case BOOLEAN:
                     case BYTE:
                     case CHAR:
                     case SHORT:
@@ -2000,10 +2044,12 @@ public class CodeGenerator {
                         emit(OpCode.D2I);
                         return;
                 }
+                break;
             }
 
             case LONG: {
                 switch (sourceKind) {
+                    case BOOLEAN:
                     case BYTE:
                     case CHAR:
                     case SHORT:
@@ -2019,10 +2065,12 @@ public class CodeGenerator {
                         emit(OpCode.D2L);
                         return;
                 }
+                break;
             }
 
             case CHAR: {
                 switch (sourceKind) {
+                    case BOOLEAN:
                     case BYTE:
                     case SHORT:
                     case INT:
@@ -2044,10 +2092,12 @@ public class CodeGenerator {
                         emit(OpCode.I2C);
                         return;
                 }
+                break;
             }
 
             case FLOAT: {
                 switch (sourceKind) {
+                    case BOOLEAN:
                     case BYTE:
                     case CHAR:
                     case SHORT:
@@ -2063,10 +2113,12 @@ public class CodeGenerator {
                         emit(OpCode.D2F);
                         return;
                 }
+                break;
             }
 
             case DOUBLE: {
                 switch (sourceKind) {
+                    case BOOLEAN:
                     case BYTE:
                     case CHAR:
                     case SHORT:
@@ -2082,6 +2134,7 @@ public class CodeGenerator {
                         emit(OpCode.F2D);
                         return;
                 }
+                break;
             }
         }
 
@@ -2210,7 +2263,7 @@ public class CodeGenerator {
             callback.emitDefault(breakTarget);
             markLabel(breakTarget);
         }
-        catch (Exception e) {
+        catch (final Exception e) {
             throw Error.codeGenerationException(e);
         }
     }
@@ -2277,7 +2330,7 @@ public class CodeGenerator {
                                     : SwitchOptions.Default);
             }
         }
-        catch (Exception e) {
+        catch (final Exception e) {
             throw Error.codeGenerationException(e);
         }
     }
@@ -2329,13 +2382,13 @@ public class CodeGenerator {
             intKeys,
             new SwitchCallback() {
                 @Override
-                public void emitCase(final int key, final Label ignore) throws Exception {
+                public void emitCase(final int key, final Label ignore) {
                     final List<String> bucket = buckets.get(key);
                     stringSwitchHelper(bucket, callback, defaultLabel, breakTarget, 0);
                 }
 
                 @Override
-                public void emitDefault(final Label breakTarget) throws Exception {
+                public void emitDefault(final Label breakTarget) {
                     emitGoto(defaultLabel);
                 }
             });
@@ -2351,7 +2404,7 @@ public class CodeGenerator {
         final StringSwitchCallback callback,
         final Label defaultLabel,
         final Label breakTarget,
-        final int index) throws Exception {
+        final int index) {
 
         final int length = bucket.get(0).length();
 
@@ -2395,7 +2448,7 @@ public class CodeGenerator {
                 }
 
                 @Override
-                public void emitDefault(final Label breakTarget) throws Exception {
+                public void emitDefault(final Label breakTarget) {
                     emitGoto(defaultLabel);
                 }
             }
@@ -2437,7 +2490,7 @@ public class CodeGenerator {
                     final List<String> bucket = buckets.get(key);
                     Label next = null;
 
-                    for (Iterator<String> it = bucket.iterator(); it.hasNext(); ) {
+                    for (final Iterator<String> it = bucket.iterator(); it.hasNext(); ) {
                         final String string = it.next();
 
                         if (next != null) {
@@ -2464,7 +2517,7 @@ public class CodeGenerator {
                 }
 
                 @Override
-                public void emitDefault(final Label breakTarget) throws Exception {
+                public void emitDefault(final Label breakTarget) {
                     pop();
                 }
             },
